"""
motifscan.cli.motif
-------------------

Command line interface for the 'motif' subcommand.
"""

import logging
import os
import shutil
import sys
from collections import defaultdict

import numpy as np

from motifscan.config import Config
from motifscan.exceptions import RemoteMotifPFMsNotFoundError, \
    MotifSetNotFoundError
from motifscan.genome import Genome
from motifscan.io.utils import merge_files
from motifscan.motif import pfms_path_fmt, MotifPwms, load_installed_pfms, \
    get_score_cutoffs
from motifscan.motif.databases import JasparDatabase
from motifscan.motif.cscore import c_score

logger = logging.getLogger(__name__)


def run(args, config_file=None):
    if args.list:
        config = Config(config_file)
        for name, _ in config.list_motif_sets():
            print(name)
        return
    if args.list_remote:
        database = JasparDatabase()
        if args.database == 'jaspar_core':
            remote_sets = database.pfms_core
            for name in remote_sets:
                print(f"{name:25}\t{database.name + '_CORE'}")
        else:
            remote_sets = database.pfms_other_collections
            for name in remote_sets:
                print(f"{name:20}\t{database.name + '_Collections'}")
        return
    if args.install:
        install_motif(args, config_file)
        return
    if args.build:
        build_motif(args, config_file)
        return
    if args.uninstall:
        uninstall_motif(args, config_file)
        return


def install_motif(args, config_file=None):
    config = Config(config_file)
    if config.has_motif_set(args.name):
        logger.error(f"Motif set {args.name!r} already exists!")
        sys.exit(1)

    motif_dir = os.path.abspath(
        args.output_dir or os.path.join(config.get_motif_dir(), args.name))
    logger.info(f"Installing motif set {args.name!r} into {motif_dir}")
    if not os.path.isdir(motif_dir):
        os.makedirs(motif_dir)
    if os.listdir(motif_dir):
        logger.error("Directory not empty! Please specify another directory "
                     "or delete files under it.")
        sys.exit(1)

    pfms_path = pfms_path_fmt.format(motif_dir, args.name)

    if args.remote:
        try:
            db = JasparDatabase()
            if args.database == 'jaspar_core':
                dst_pfms = db.download_core(args.remote, motif_dir)
            else:
                dst_pfms = db.download_other_collections(args.remote,
                                                         motif_dir)
            logger.debug(
                f"Renaming downloaded file to {os.path.basename(pfms_path)}")
            shutil.move(dst_pfms, pfms_path)
        except RemoteMotifPFMsNotFoundError as e:
            logger.error(e)
            sys.exit(1)
    else:
        logger.info("Copying the PFMs file(s)")
        merge_files(args.pfm_files, pfms_path)

    logger.info("Updating the config file")
    config.set_motif_path(args.name, motif_dir)
    config.write()
    logger.info("Successfully installed!")
    if args.genome:
        build_motif(args, config_file)


def build_motif(args, config_file=None):
    if args.build:
        name = args.build
    else:
        name = args.name
    logger.info(
        f"Building motif set {name!r} for genome assembly {args.genome!r}")
    genome = Genome(args.genome)
    pfms = load_installed_pfms(name)

    logger.info("Converting motif PFMs to PWMs")
    max_length = 0
    pwms = MotifPwms(name=name, genome=genome.name)
    for pfm in pfms:
        max_length = max(max_length, pfm.length)
        pwm = pfm.to_ppm().to_pwm(genome.bg_freq)
        pwms.append(pwm)

    cutoffs_all = []
    for i in range(args.n_repeat):
        if args.n_repeat > 1:
            logger.info(
                f"Building motif score cutoffs: {i + 1} of {args.n_repeat}")
        if args.seed is not None:
            seed = args.seed + i
        else:
            seed = None
        logger.info("Random sampling background sequences")
        seqs = list(genome.random_sequences(args.n_random, max_length,
                                            args.max_n, seed))

        logger.info("Calculating background motif scores")
        matrices = [pwm.matrix.tolist() for pwm in pwms]
        sampling_scores = c_score(matrices, seqs, 3, args.n_threads)

        logger.info("Getting motif score cutoffs")
        cutoffs_all.append(get_score_cutoffs(sampling_scores))

    if args.n_repeat > 1:
        logger.info("Saving averaged motif score cutoffs")
    else:
        logger.info("Saving motif score cutoffs")

    for i, pwm in enumerate(pwms):
        cutoffs = defaultdict(list)
        for pwms_cutoffs in cutoffs_all:
            pwm_cutoffs = pwms_cutoffs[i]
            for p_value, cutoff in pwm_cutoffs.items():
                cutoffs[p_value].append(cutoff)
        for p_value in cutoffs:
            cutoff = np.mean(cutoffs[p_value])
            cutoff = np.around(cutoff, 8)
            pwm.set_cutoff(p_value=p_value, cutoff=cutoff)
    pwms.save_built_pwms()
    logger.info("Successfully built!")


def uninstall_motif(args, config_file=None):
    try:
        config = Config(config_file)
        path = config.get_motif_path(args.uninstall)
    except MotifSetNotFoundError as e:
        logger.error(e)
        sys.exit(1)

    logger.info(f"Uninstalling motif set {args.uninstall!r}")
    if os.path.isdir(path):
        logger.info(f"Removing files under {path}")
        try:
            shutil.rmtree(path)
        except Exception as e:
            logger.error(f"Failed to remove the motif directory: {e}")
            sys.exit(1)

    logger.info("Updating the config file")
    config.remove_motif_path(args.uninstall)
    config.write()
    logger.info("Successfully uninstalled!")



"""
motifscan.scanner
-----------------

Module for core motif scanner.
"""

import logging
import os
from collections import namedtuple

from motifscan.motif.cscore import c_scan_motif

logger = logging.getLogger(__name__)

MotifSite = namedtuple("MotifSite", ['start', 'score', 'strand'])


class Scanner:
    """
    Motif scanner for a list of genomic regions.

    Parameters
    ----------
    genome : `motifscan.genome.Genome`
        A motifscan genome object to extract sequences from.
    regions : list of `GenomicRegion`
        Specifies the genomic regions for scanning.
    window_size : int, optional
        If `window_size` <= 0, then the whole input region will be scanned.
        Otherwise, only scans the the specified window centered at the summit
        or midpoint of each genomic region.
    strand : {'both', '+', '-'}, optional
        Which strand is used to find motif occurrences, default='both'.
    p_value : str, optional
        Significance level for the motif score cutoff, default='1e-4'.
    remove_dup ：bool, optional
        If True (default), remove adjacent duplicated motif occurrences whose
        distance is less than the motif length.
    n_threads : int, optional
        Number of threads used in parallel, default=1.
    """

    def __init__(self, genome, regions, window_size=0, strand='both',
                 p_value='1e-4', remove_dup=True, n_threads=1):
        if window_size <= 0:
            self.window_size = 0
        else:
            self.window_size = window_size
        self.extend = window_size // 2
        if strand in ['both', '+', '-']:
            self.strand = strand
        else:
            raise ValueError(f"invalid strand option: {strand!r}")
        self.p_value = p_value
        self.remove_dup = remove_dup
        n_threads = int(n_threads)
        n_cpu = os.cpu_count()
        if n_threads > n_cpu:
            logger.warning(f"Threads number exceed the number of CPUs, "
                           f"using {n_cpu} instead")
            n_threads = n_cpu
        if n_threads < 1:
            n_threads = 1
        self.n_threads = n_threads
        self.seq_starts = []
        self.seq_ends = []
        self.sequences = []
        self._extract_seq(genome=genome, regions=regions)

    def _extract_seq(self, genome, regions):
        """Extract the sequences on the forward strand and record the
        underlying coordinates.
        """
        logger.debug("Extracting sequences")
        for region in regions:
            if self.window_size <= 0:
                seq_start = region.start
                seq_end = region.end
            else:
                seq_start = max(region.summit - self.extend, 0)
                seq_end = min(region.summit + self.extend,
                              genome.chrom_sizes[region.chrom])
            self.seq_starts.append(seq_start)
            self.seq_ends.append(seq_end)
            self.sequences.append(
                genome.fetch_sequence(region.chrom, seq_start, seq_end))

    def scan_motifs(self, pwms):
        """Scan for motif occurrences given the motif PWMs.

        Parameters
        ----------
        pwms : `motifscan.motif.MotifPwms`
            Motif Pwms to be scanned.

        Returns
        -------
        motif_sites : nested list, (n_pwms, n_regions, )
            Identified motif occurrences.
        """
        # check if the corresponding motif score cutoff is set
        for pwm in pwms:
            try:
                pwm.cutoffs[self.p_value]
            except (TypeError, KeyError):
                raise ValueError(
                    f"PWM has no motif score cutoff set for P-value "
                    f"{self.p_value!r}")

        logger.debug(f"Scanning motif PWMs")
        matrices = []
        cutoffs = []
        for pwm in pwms:
            matrices.append(pwm.matrix.tolist())
            cutoffs.append(pwm.cutoffs[self.p_value])

        if self.strand == '+':
            strand_arg = 1
        elif self.strand == '-':
            strand_arg = 2
        else:
            strand_arg = 3

        sites = c_scan_motif(matrices, cutoffs, self.sequences, strand_arg,
                             self.n_threads)

        motif_sites = make_motif_sites(sites, self.seq_starts)
        if self.remove_dup:
            lengths = [pwm.length for pwm in pwms]
            motif_sites = deduplicate_motif_sites(motif_sites, lengths)
        return motif_sites


def make_motif_sites(sites, seq_starts):
    """Given the pooled motif sites from C extension, rearrange motif sites by
    region(sequence), returns a nested list of shape (n_pwms, n_seqs, n_sites).
    """
    motif_sites = []
    for pwm_idx, sites_pwm in enumerate(sites):
        motif_sites.append([])
        for _ in seq_starts:
            motif_sites[pwm_idx].append([])
        for site in sites_pwm:
            seq_idx, pos_idx, score, strand = site
            start = seq_starts[seq_idx] + pos_idx
            if strand == 1:
                strand = '+'
            else:
                strand = '-'
            motif_sites[pwm_idx][seq_idx].append(
                MotifSite(start=start, score=score, strand=strand))
    return motif_sites


def _deduplicate_sites(sites, length):
    idx = 0
    if len(sites) > 1:
        while idx + 1 < len(sites):
            site_curr = sites[idx]
            site_next = sites[idx + 1]
            if site_next.start - site_curr.start < length:
                if site_curr.score >= site_next.score:
                    sites.pop(idx + 1)
                else:
                    sites.pop(idx)
            else:
                idx += 1


def deduplicate_motif_sites(motif_sites, lengths):
    """Deduplicate adjacent motif occurrences with distance less than the motif
    length and only keep the one with higher motif score.
    """
    motif_sites_dedup = []
    for sites_pwm, length in zip(motif_sites, lengths):
        sites_pwm_dedup = []
        for sites in sites_pwm:
            # split fwd sites and rev sites
            sites_fwd = []
            sites_rev = []
            for site in sites:
                if site.strand == '+':
                    sites_fwd.append(site)
                else:
                    sites_rev.append(site)
            _deduplicate_sites(sites_fwd, length)
            _deduplicate_sites(sites_rev, length)
            sites_dedup = sites_fwd + sites_rev
            sites_dedup.sort(key=lambda x: x.start)
            sites_pwm_dedup.append(sites_dedup)
        motif_sites_dedup.append(sites_pwm_dedup)
    return motif_sites_dedup
